/*   Apres avoir passé pas mal d'heures à comprendre pourquoi je ne comprenais pas !!
     Je me suis enfin rendu compte que les processeurs AVR possédaient des registres
     particulier pour stocker les points de retour des procédures appelantes.
     Contrairement aux processeurs "standards" style Intel et/ou Motorola dans lesquels
     la pile constitue le coeur du systeme de fonctionnement des appels de procedures,
     dans les AVR(s), celle ci ne sert pas (habituellement) à stocker l'adresse de retour
     Donc les macros développés (langage inline) sont similaires aux procedures de même nom.
     Il n'y a pas lieu de s'inquiéter des octets "pompés" par les push(s) classiques des adersses
     de retour.

     Pour mieux comprendre comment C++ stocke ses données et les mécanismes implicites et explicites
     à mettre en oeuvre pour conserver une mémoire "sans trou" voici un petit programme
     qui bien qu'il ne fasse pas grand chose exploite les procédures de la bibliothèque
     de visualisation de la mémoire et, en prime constitue un exemple de structure dynamique chainée
     via la classe TableauDynamique

     Utilisez ce programme, rajoutez des visualisation, des variables ...
     et surtout, prenez une bonne dose d'aspirine avant de vous lancer...
     
     Ce programme appuie sur les procédures de la bibliotheque MemoryUsage V2.30

    Yves Cordier
    07/04/2020
*/
#include <MemoryUsage.h>

// Une structure objet chainée
//============================
class TableauDynamique
{
  private : // les attributs privés. Leurs acces se font via des méthodes
    float contenu;
    int ordre = 0; 
    TableauDynamique *suiv;
  public :
    // Costructeurs et destructeurs
    // *****************************

    // Constructeur de tableau "i cases" de 0 à i-1
    TableauDynamique(int i)
    { contenu = 0.0;
      if (i > 1) suiv = new TableauDynamique(i - 1);
      else  suiv = NULL;
      ordre = i - 1;
    }
    // constructeur tableau 1 case
    TableauDynamique()
    {
      contenu = 0;
      suiv = NULL;
      ordre = 0;
    }
    // destructeur du tableau
    ~TableauDynamique()
    {
      if (suiv != NULL)
      {
        delete suiv; // invoque le destructeur de suiv
      }
    }
    // **********************************
    // Gestion globale du tableau
    // **********************************

    // Ajoute une case en tête de tableau => augmenter sa taille de 1
    void Ajoute(float f = 0)
    { TableauDynamique *T = new TableauDynamique();
      T->ordre = ordre;
      T->contenu = contenu;
      ordre++;
      contenu = f;
      T->suiv = suiv;
      suiv = T;
    }

    // enleve la premiere case => diminuer sa taille de 1
    void Enleve()
    {
      TableauDynamique *T = suiv;
      if (suiv)
      { suiv = suiv->suiv;
        delete T;
      }
    }

    // recalibre un tableau (le mettre à une taille définie)
    void NewSize(int n)
    {
      while (n > (ordre + 1))  Ajoute();
      while (n < (ordre + 1)) Enleve();
    }

    // Gestion des valeurs
    // *******************
    // remplissage global
    void Remplissage(float f)
    {
      TableauDynamique *T = suiv;
      while (T) {
        T->contenu = f;
        T = T->suiv;
      }
      contenu = f;
    }

    // remplissage et lecture des cases
    float getVal(int i)
    { if (i == ordre) return contenu;
      else if (suiv) return suiv->getVal(i);
      return 0; // par securité on a depassé
    }
    void setVal(float f, int i)
    { if (i == ordre) contenu = f;
      else
      {
        if (suiv) suiv->setVal(f, i);
      }
    }
    // recherche de la taille
    // version non recursive
    int TailleNR()
    { TableauDynamique *T = suiv;
      int i = 1;
      while (T) {
        T = T->suiv;
        i++;
      }
      return i;
    }
    // version recursive
    int TailleRec()
    { if (suiv) return suiv->TailleRec() + 1;
      return 1;
    }
    // version futée comme les cases sont stockées en ordre inverse !!
    int Taille()
    {
      return (ordre + 1);
    }


    // renvoi l'adresse de la case d'ordre i
    void* Adresse(int i)
    {
      if (ordre == i) return this; // this est le pointeur sur l'objet
      if (suiv) return suiv->Adresse(i);
      return NULL; // pas de cases 
    }

    // procedures diverses
    // *******************
    void SerialPrintOrdreMontant()
    {
      if (suiv) suiv->SerialPrintOrdreMontant();
      printCase();
    }
    
    void SerialPrintOrdreDecroissant()
    {
      printCase();
      if (suiv) suiv->SerialPrintOrdreDecroissant();
    }

    void printCase()
    {
      Serial.print('[');
      Serial.print(ordre);
      Serial.print("] : ");
      Serial.println(contenu);
    }
};

// Fin de definition de la classe
// ===============================


// Une preocedure de demo
// ======================
void proceduredemo()
{
  Serial.println("début de procedure demo");
  SRamDisplay();// function
  //_SRamDisplay(); // idem en macro

  TableauDynamique T1 = TableauDynamique(10); // Creation d'un tableau dynamique local de 10 cases
  // le premier element est dans la pile, les autres dans le tas
  // on se rapelle qu'a la création, le tableau est stocké à l'envers !!
  Serial.print("localisation case N°9 de T1 "); WhereIsThisAdress(T1.Adresse(9));
  Serial.print("localisation case N°0 de T1 "); WhereIsThisAdress(T1.Adresse(0));

  // Un autre tableau créé sous forme de pointeur
  TableauDynamique *T2 = new TableauDynamique(10); // Creation d'un pointeur sur un tableau dynamique
  Serial.println("apres creation T1 puis T2");
  SRamDisplay();// seul le pointeur est dans la pile le tableau entier lui est dans le tas
  Serial.print("localisation case N°9 de T2"); WhereIsThisAdress(T2->Adresse(9)); // function
  Serial.println();
  Serial.println("On travail on travail !!!!");
  // On fait plein de chose avec nos données
  for (int i = 0; i < T2->Taille(); i++)
  {
    T2->setVal(i, i);
  }
  T2->SerialPrintOrdreMontant();

  Serial.println();
  Serial.println("Repos!!");
  Serial.println();
  delete T2;// on détruits les parties logées dans le tas (toutes)

  Serial.println("apres destruction de T2");
  SRamDisplay();

  // T1 étant une variable locale (créé dans la fonction) de type objet, le compilateur avant de quitter la
  // fonction invoque la fonction delete de cet objet qui apelle le destructor
  // que nous avons soigneusement écrit pour qu'il vide Toutes les composantes
  // de l'objet.
  // la partie logée dans la pile sera détruite automatiquement en sortie de procedure
  // par incrementation du SP.
}

// *************************************************************************
// Le setup Standard
// *************************************************************************
void setup()
{
  // put your setup code here, to run once:
  Serial.begin(9600);
  Serial.println("debut setup");
  SRamDisplay();
  char S[] = "hello world0123456789"; // stockée dans la pile donc automatiquement détruit en sortie de setup
  byte i = 0;
  while (!S[i]) i++;
  Serial.println("apres declaration S");
  String Sst1 = "bonjour";
  Serial.print("Variable Sst1 locale de type String partie principale créé dans :"); WhereIsThisAdress(&Sst1);
  Serial.print("Variable Sst1 de type objet stockage de ses données dans "); WhereIsThisAdress(Sst1.c_str());

  String *Sst2 = new String("re bonjour"); // Creation d'un pointeur sur la variable
  Serial.print("Variable Sst2 locale de type pointeur sur une String créée dans :"); WhereIsThisAdress(&Sst2);
  Serial.print("Variable *Sst2 de type objet String stockage de ses données dans "); WhereIsThisAdress(Sst2);
  Serial.print("Variable Sst2 de type objet stockage de ses données dans "); WhereIsThisAdress(Sst2->c_str());

  SRamDisplay();
  delete Sst2; // il nous appartient de la détruire. Le pointeur lui sera détruit automatiquement
  SRamDisplay();
  // Sst1 sera détrite autolmatiquement en sortie de setup de par son destructeur (intégré)
}
// ********************************************************************************************************
// Déclarations de variable globales
// ********************************************************************************************************
// Deux variables standard
float f1; // stockée dans la partie BSS car non initialisée
float f2 = PI; // stockée dans la partie Data car initialisée

// ********************************************************************************************************
// Procedure loop standard
// ********************************************************************************************************

void loop()
{
  Serial.println("entree dans loop");
  Serial.print("Ou est f1 ? ");
  WhereIsThisAdress(&f1); // nota le prefix & devant une variable permet de renvoyer l'adresse (donc une valeur de pointeur !!)
  Serial.println("Montre nous ou se loge f2");
  PosVarSRam(&f2);
  proceduredemo();
  Serial.println("apres être sortie de la procedure de demo");
  SRamDisplay();
  while (true) {} // on attends !!
}

// ********************************************************************************************************
// Pour rappel allure de la procedure main intégrée (en simplifié !!)
/* 
 *  void main()
 *  {
 *  setup()
 *  while(true) loop();
 *  }
 */
// ********************************************************************************************************
